use std::{any::Any, cell::RefCell, rc::Rc};

use clay_layout::{
    Clay,
    bindings::Clay_TextElementConfig,
    elements::{
        CornerRadius,
        rectangle::Rectangle,
        text::{Text, TextElementConfig},
    },
    fixed,
    id::Id,
    layout::Layout,
    math::Dimensions,
};

use font_kit::{family_name::FamilyName, properties::Properties, source::SystemSource};
use gui::GUI;
use minifb::{Window, WindowOptions};
use raqote::{DrawOptions, DrawTarget, SolidSource};

mod gui;

fn main() {
    let clay = Clay::new((800., 600.).into());

    let mut window = Window::new("Raqote", 800, 40, WindowOptions {
        borderless: true,
        scale_mode: minifb::ScaleMode::Center,
        resize: true,
        ..Default::default()
    })
    .unwrap();

    let font = SystemSource::new()
        .select_best_match(&[FamilyName::Monospace], &Properties::new())
        .unwrap()
        .load()
        .unwrap();

    window.set_target_fps(60);

    let mut size = window.get_size();
    let mut dt = DrawTarget::new(size.0 as i32, size.1 as i32);

    clay.measure_text_function(|text, config| {
        let config: Clay_TextElementConfig = unsafe {
            *std::mem::transmute::<TextElementConfig, *mut Clay_TextElementConfig>(config)
        };

        Dimensions {
            width: config.fontSize as f32,
            height: config.fontSize as f32 * text.len() as f32,
        }
    });

    loop {
        dt.clear(SolidSource::from_unpremultiplied_argb(
            0xff, 0xff, 0xff, 0xff,
        ));
        clay.begin();

        // Adds a red rectangle with a corner radius of 5.
        // The Layout makes the rectangle have a width and height of 50.
        clay.with(
            [
                Id::new("red_rectangle"),
                Layout::new().width(fixed!(50.)).height(fixed!(50.)).end(),
                Rectangle::new()
                    .color((0xFF, 0x00, 0x00).into())
                    .corner_radius(CornerRadius::All(5.))
                    .end(),
            ],
            |c| {
                c.with(
                    [
                        Id::new("test_text"),
                        Layout::new().end(),
                        Rectangle::new().color((0xFF, 0xFF, 0xFF).into()).end(),
                        // Text::new().color((0xFF, 0xFF, 0xFF).into()).end()
                    ],
                    |_| {},
                );
            },
        );

        // Return the list of render commands of your layout
        let render_commands = clay.end();

        for command in render_commands {
            // println!("Id of the element: {}", command.id); // Note: Ids are in fact numbers generated by Clay
            // println!("Bounding box: {:?}", command.bounding_box);
            // println!("Type and config: {:?}", command.config);

            match command.config {
                clay_layout::render_commands::RenderCommandConfig::Rectangle(rectangle) => {
                    dt.fill_rect(
                        command.bounding_box.x,
                        command.bounding_box.y,
                        command.bounding_box.width,
                        command.bounding_box.height,
                        &raqote::Source::Solid(SolidSource {
                            r: (rectangle.color.r * 255.0) as u8,
                            g: (rectangle.color.g * 255.0) as u8,
                            b: (rectangle.color.b * 255.0) as u8,
                            a: (rectangle.color.a * 255.0) as u8,
                        }),
                        &DrawOptions::new(),
                    );
                }
                clay_layout::render_commands::RenderCommandConfig::Border(border_container) => {
                    todo!()
                }
                clay_layout::render_commands::RenderCommandConfig::Text(_, text) => todo!(),
                clay_layout::render_commands::RenderCommandConfig::Image(image) => todo!(),
                clay_layout::render_commands::RenderCommandConfig::ScissorStart() => todo!(),
                clay_layout::render_commands::RenderCommandConfig::ScissorEnd() => todo!(),
                clay_layout::render_commands::RenderCommandConfig::Custom(custom) => todo!(),
                clay_layout::render_commands::RenderCommandConfig::None() => {}
            }
        }

        if window.is_key_down(minifb::Key::Escape) {
            break;
        }

        let (width, height) = window.get_size();

        if width as i32 != dt.width() || height as i32 != dt.height() {
            dt = DrawTarget::new(width as i32, height as i32);
            clay.layout_dimensions((width as f32, height as f32).into());
            size = (width, height);
            // window.update();
        } else {
            window
                .update_with_buffer(dt.get_data(), size.0, size.1)
                .unwrap();
        }
    }
}
